sr      =  44100
kr      =  4410
ksmps   =  10
nchnls  =  1

; Sawtooth wave (also used for PWM).
itmp	ftgen 1, 0, 16384, 7, 1, 16384, -1
; The sawtooth-triangle ramp can be generated by integrating
; a pulse-width modulated square wave (which is the difference of
; two sawtooth waves with different phase); however, it is more
; efficient (and also eliminates problems related to the use of
; a leaky integrator in the instruments) to start from an already
; integrated waveform. The sawtooth wave (x is in the range 0 to 1)
; is:
;   y = 1 - 2*x
; after integrating this we get:
;   y = x - x^2
; GEN03 generates the polynomial (with normalization, the function
; is actually 4*(x - x^2), as the maximum value was originally 0.25
; at x = 0.5)
itmp	ftgen 2, 0, 16384, 3, 0, 1, 0, 1, -1
; Sine (to be used by LFOs). Read with interpolation, so a small
; table is sufficient.
itmp	ftgen 5, 0, 256, 10, 1
; Window for 1/16 overlap (used by "soft sync"). If the total length
; is 16384 samples, the lengths of the three segments are (note that
; for an overlap of 1/16 we actually have to divide table length by
; 17, and not 16):
;   16384 *  (1/17) =   963.7647
;   16384 * (15/17) = 14456.4706
;   16384 *  (1/17) =   963.7647
itmp	ftgen 6, 0, 16384, 7, 0, 964, 1, 14456, 1, 964, 0

#include "fgen.h.orc"

; Generate bandlimited waveforms using FGEN_128 macro.

$FGEN_128(l01'100'4096'1'1'sr/2)
$FGEN_128(l02'300'4096'2'1'sr/2)

#include "oscillators.h.orc"

/* ---- instr 1: sawtooth wave with i-rate table ---- */

	instr 1

; For many cases, it is sufficient to use i-rate table numbers,
; which allows using the (faster) oscili opcode instead of
; phasor + tableikt. To get a "clean" bandwidth of about 16000 Hz
; with sr = 44100 Hz, the minimum and maximum transpose factors (for
; e.g. vibrato or pitch bend) are:
;   16000 / 22050	    = 0.7256	(-5.5 semitones)
;   (44100 - 16000) / 22050 = 1.2744	(+4.2 semitones)

icps	=  440		; base frequency

; The table number can be calculated from the oscillator frequency
; with the following formula:
;   (base ftable) + 69 + 12 * (log(frequency / 440) / log (2))
; this is rounded to the nearest integer by adding 0.5 and using
; int(). "base ftable" is 100 for sawtooth wave.

kfnum	Cps2Fnum icps, 100

; add some variation to the frequency within the allowed limits

ktrans	lfo 0.25, 1 / p3, 0
ktrans	=  ktrans + 1		; range: 0.75 to 1.25

; oscillator

a1	oscili 20000, icps * ktrans, i(kfnum)

	out a1

	endin

/* ---- instr 2: sawtooth wave with k-rate table ---- */

	instr 2

; This instrument is similar to instr 1, but uses a k-rate table
; number to allow more variation in frequency (however, at high
; oscillator frequency, the switching of tables may result in
; clicks; this is audible if frequency is higher than about 2 kHz).
; This also means that the table read opcode has to support k-rate
; table number; some of such units:
;   tablekt    (not recommended as the lack of interpolation
;               reduces quality)
;   tableikt
;   tablexkt   (faster than tableikt but has more parameters)
;   oscilikt   (oscili with k-rate table number)
;   grain2/3   (these are useful for generating more complex
;   oscbnk      sounds)

; oscillator frequency
kfrq	expon 50, p3, 3200
; frequency -> table number
kfn	Cps2Fnum kfrq, 100
; oscillator
a1	oscilikt 1, kfrq, kfn

	out a1 * 20000

	endin

/* ---- instr 3: PWM ---- */

	instr 3

; frequency envelope
kfrq	expon 200, p3, 400
; frequency -> table number
kfn	Cps2Fnum kfrq, 100
; LFO for pulse width
kpw	oscili 0.45, 0.8, 5, 0
kpw	=  kpw + 0.5	; 0.05 to 0.95
; oscillator
a1	PWMOsc 20000, kfrq, kfn, kpw, 0

	out a1

	endin

/* ---- instr 4: sawtooth-triangle morph ---- */

	instr 4

; frequency envelope
kfrq	expon 200, p3, 400
; frequency -> table number
; use 4x(1-x) waveform (i.e. integrated sawtooth)
kfn     Cps2Fnum kfrq, 300
; kpw = pulse width (0.01 - 0.99). 0 and 1 are not allowed as
; these values would result in division by zero.
; 0.01: sawtooth down, 0.5: triangle, 0.99: sawtooth up
kpw	oscili 0.45, 0.8, 5, 0
kpw	=  kpw + 0.5	; 0.05 to 0.95
; oscillator
a1	SawTriOsc 20000, kfrq, kfn, kpw, 0

	out a1

	endin

/* ---- instr 5: granular "soft sync" with PWM ---- */

	instr 5

; This instrument is based on instr 3 (PWM), but grain3 is used to
; generate overlapping windows.

; sync frequency
kfrqs	expon 100, p3, 200
; oscillator frequency
kfrq	expon 1000, p3, 500
; frequency -> table number
kfn     Cps2Fnum kfrq, 100

; k2 = pulse width (0 - 1)
k2	oscili 0.45, 0.6, 5, 0
k2	=  k2 + 0.5	; 0.05 to 0.95

; granular synthesis parameters:
;   density =         sync ("master") frequency
;   grain frequency = oscillator ("slave") frequency
;   grain duration =  (1 + 1/16) / density    (to get 1/16 overlap)
; imode is set to 2 to get grain parameters continuously controlled
; by phase and frequency, instead of the default of keeping the
; settings the grain was launched with.
; Randomization is not used, so all parameters related to it are 0.
a01	grain3 kfrq, 0, 0, 0, 1.0625 / kfrqs, kfrqs, 2, kfn, 6, \
	       0, 0, 0, 2
a02	grain3 kfrq, 1 - k2, 0, 0, 1.0625 / kfrqs, kfrqs, 2, \
	       kfn, 6, 0, 0, 0, 2
a1	=  a01 - a02
; correct DC offset (see instr 3)
aofs	interp 2 * k2 - 1, 0, 1
	vincr a1, aofs

	out a1 * 20000

	endin

/* ---- instr 6: hard sync using oscilikts ---- */

	instr 6

; kfrq: frequency of "slave" oscillator
kfrq	expon 400, p3, 1200
; kphs: phase (0-1) to which the oscillator is re-initialized
kphs	line 0.1, p3, 0.9

; two "master" oscillators running at a different frequency
; re-initialization occurs when any of these oscillators is
; at phase=0
atmp1	phasor 100
atmp2	phasor 150
; oscilikts reinitializes on positive sync input; this can be
; achieved by inverting and differentiating the outputs of
; the phasors
async	diff 1 - (atmp1 + atmp2)

; the usual frequency to table number conversion
kfn     Cps2Fnum kfrq, 100
; oscillators
a1	oscilikts 1, kfrq, kfn, async, 0
a2	oscilikts 1, kfrq, kfn, async, -kphs
; mix outputs with DC correction (see instr 3) for PWM
; (not very accurate this time, as pulse width is set only at
; re-initialization)
aofs	interp 2 * kphs - 1, 0, 1
a1	=  (aofs + a1 - a2) * 20000

	out a1

	endin

